import asyncio
import json
import os
import uuid
from typing import Any, Dict, List

from openai import OpenAI
from playwright.async_api import Page, async_playwright
from tarsier import GoogleVisionOCRService, Tarsier

from bananalyzer.data.fetch_schemas import get_fetch_schema
from bananalyzer.data.schemas import Eval, Example


async def add_examples_to_json(
    urls: List[str],
    schema: Dict[str, Any],
    metadata: Dict[str, Any],
    tarsier_client: Tarsier,
    openai_client: OpenAI,
) -> None:
    for url in urls:
        example = await generate_fetch_example(
            url, schema, metadata, tarsier_client, openai_client
        )
        add_example_to_json(example)


async def generate_fetch_example(
    url: str,
    schema: Dict[str, Any],
    metadata: Dict[str, Any],
    tarsier_client: Tarsier,
    openai_client: OpenAI,
) -> Example:
    example_id = await download_as_mhtml(url)
    mhtml_path = os.path.abspath(f"./static/{example_id}/index.mhtml")
    url = f"file://{mhtml_path}"
    eval_expected = await llm_annotate_example(
        url, schema, tarsier_client, openai_client
    )
    eval = Eval(type="json_match", expected=eval_expected)
    example = Example(
        id=example_id,
        url=url,
        source="mhtml",
        category=metadata["category"],
        subcategory=metadata["subcategory"],
        type="fetch",
        fetch_id=metadata["fetch_id"],
        evals=[eval],
        goal=None,
    )
    if hasattr(example, "goal"):
        delattr(example, "goal")
    return example


async def llm_annotate_example(
    url: str, schema: Dict[str, Any], tarsier_client: Tarsier, openai_client: OpenAI
) -> Dict[str, Any]:
    async with async_playwright() as playwright:
        browser = await playwright.chromium.launch(headless=False)
        context = await browser.new_context()
        page = await context.new_page()
        await page.goto(url)
        await page.wait_for_load_state("domcontentloaded")
        page_text, _ = await tarsier_client.page_to_text(page, False)

    prompt = f"""Here's an OCR'd screenshot of the details page for a listing on a website.
{page_text}

Here is the schema we want to map information from the details page into. If an attribute in this schema doesn't have corresponding information on the details page, do NOT include that attribute in your output. If there is no corresponding information on the page at all, return an empty JSON.
```json
{schema}
```

For each attribute in the schema, find information on the details page that would serve as its value and return it as part of a JSON object (formatted as ```json ...``` with double quotes for property names) mapping the attribute key to the information. For attributes that are a list of objects, remember to provide well-formed JSON for each object in the list."""

    completion = openai_client.chat.completions.create(
        model="gpt-4",
        messages=[{"role": "user", "content": prompt}],
    )
    output = completion.choices[0].message.content
    assert isinstance(output, str)

    json_start = output.find("```json")
    if json_start == -1:
        details = {}
    else:
        json_start += len("```json")
        json_end = output.find("```", json_start)
        json_output = output[json_start:json_end].strip()
        try:
            details = json.loads(json_output)
        except json.JSONDecodeError:
            details = {}

    return details


async def download_as_mhtml(url: str) -> str:
    async with async_playwright() as playwright:
        browser = await playwright.chromium.launch(headless=False)
        context = await browser.new_context()
        page = await context.new_page()
        client = await context.new_cdp_session(page)

        await page.goto(url)

        # Capture page content as MHTML
        result = await client.send("Page.captureSnapshot", {"format": "mhtml"})
        mhtml = result["data"]

        # Write MHTML content to the specified file
        id = str(uuid.uuid4())
        folder_path = f"./static/{id}"
        os.makedirs(folder_path, exist_ok=True)
        file_path = os.path.join(folder_path, "index.mhtml")
        with open(file_path, "w") as f:
            f.write(mhtml)

        return id


def add_example_to_json(example: Example) -> None:
    json_file_path = "./static/examples.json"
    with open(json_file_path, "r") as json_file:
        data = json.load(json_file)

    data.append(example.dict())

    with open(json_file_path, "w") as json_file:
        json.dump(data, json_file, indent=4)
        json_file.write("\n")


async def main() -> None:
    openai_client = OpenAI()
    ocr_service = GoogleVisionOCRService({})
    tarsier_client = Tarsier(ocr_service)

    urls = [
        "https://www.skadden.com/professionals/l/lanstra-allen-l",
        "https://www.sullcrom.com/Lawyers/Erik-D-Lindauer",
        "https://www.winston.com/en/professionals/padmanabhan-krishnan",
        "https://www.mwe.com/people/michael-abercrombie/",
    ]
    schema = get_fetch_schema("attorney").model_fields
    metadata = {
        "category": "legal",
        "subcategory": "attorney_details",
        "fetch_id": "attorney",
    }
    await add_examples_to_json(urls, schema, metadata, tarsier_client, openai_client)


if __name__ == "__main__":
    asyncio.run(main())



import os
import tempfile
from pathlib import Path
from typing import Awaitable, Callable, List, Tuple

import pytest
from pydantic import BaseModel

from bananalyzer.data.schemas import Example
from bananalyzer.hooks import BananalyzerPytestPlugin
from bananalyzer.schema import AgentRunnerClass, PytestArgs

TestType = Callable[[], Awaitable[None]]


class BananalyzerTest(BaseModel):
    code: str
    example: Example


def create_test_file(
    tests: List[BananalyzerTest],
    prefix: str,
    cache_dir: Path,
    runner: AgentRunnerClass,
    headless: bool,
    single_browser_instance: bool,
) -> str:
    with tempfile.NamedTemporaryFile(
        mode="w+",
        delete=False,
        prefix=prefix,
        suffix=".py",
        dir=cache_dir,
    ) as f:
        # noinspection PyUnresolvedReferences
        f.write(
            f"""
import pytest
import pytest_asyncio
import asyncio

from bananalyzer.data.examples import get_example_by_url
from playwright.async_api import async_playwright


@pytest.fixture(scope="session")
def event_loop():
    loop = asyncio.new_event_loop()
    yield loop
    loop.close()


@pytest.fixture(scope="session")
def agent():
    import importlib.util
    import sys
    from pathlib import Path
    import pprint


    path = Path("{runner.class_path}")
    module_name = path.stem
    spec = importlib.util.spec_from_file_location(module_name, str(path))
    module = importlib.util.module_from_spec(spec)

    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    print(f"Loaded agent module", module)

    agent_ = getattr(module, '{runner.class_name}')()
    yield agent_


@pytest_asyncio.fixture(scope="{'session' if single_browser_instance else 'class'}")
async def page():
    async with async_playwright() as p:
        browser = await p.chromium.launch(headless={headless})
        context = await browser.new_context(viewport={{ 'width': 1280, 'height': 1024 }})
        page = await context.new_page()
        yield page
        await browser.close()

"""
        )
        for test_content in tests:
            f.write(f"{test_content.code}\n\n")

    return f.name


def run_tests(
    tests: List[BananalyzerTest],
    runner: AgentRunnerClass,
    pytest_args: PytestArgs,
    headless: bool = False,
    single_browser_instance: bool = False,
) -> Tuple[int, Path]:
    """
    Create temporary test files based on intent, run them, and then delete them
    """
    intents = {test.example.type for test in tests}
    intent_separated_tests = [
        [test for test in tests if test.example.type == intent] for intent in intents
    ]

    cache_dir = Path(os.getcwd()) / ".banana_cache"
    cache_dir.mkdir(exist_ok=True)
    with open(cache_dir / ".gitignore", "w") as f:
        f.write("# Generated by bananalyzer automatically\n*")

    with tempfile.TemporaryDirectory(dir=cache_dir) as temp_dir:
        temp_path = Path(temp_dir)

        test_file_names = [
            create_test_file(
                tests,
                f"{tests[0].example.type}_intent_",
                temp_path,
                runner,
                headless,
                single_browser_instance,
            )
            for tests in intent_separated_tests
        ]

        report_path = cache_dir / f"{temp_path.stem}_report.html"
        args = (
            test_file_names
            + (["-s"] if pytest_args.s else [])
            + ([f"-n {pytest_args.n}"] if pytest_args.n else [])
            + (["-q"] if pytest_args.q else ["-vvv"])
            + [f"--junitxml={str(report_path)}"]
        )

        return pytest.main(args, plugins=[(BananalyzerPytestPlugin())]), report_path



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Union

from playwright.async_api import Page

from bananalyzer.data.schemas import Example

AgentResult = Union[Dict[str, Any], List[Dict[str, Any]], List[str], str]


class AgentRunner(ABC):
    """
    Wrapper class clients must implement to run an agent against the evaluations
    """

    @abstractmethod
    async def run(
        self,
        page: Page,
        eval_context: Example,
    ) -> AgentResult:
        pass



from typing import List, Optional

from pydantic import BaseModel, Field

from bananalyzer.data.schemas import GoalType


class PytestArgs(BaseModel):
    s: bool
    n: Optional[int]
    q: bool


class Args(BaseModel):
    path: str
    headless: bool
    single_browser_instance: bool
    id: Optional[str] = Field(default=None)
    domain: Optional[str] = Field(default=None)
    intent: Optional[GoalType] = Field(default=None)
    category: Optional[str] = Field(default=None)
    subcategory: Optional[str] = Field(default=None)
    skip: List[str]
    type: Optional[str] = Field(default=None)
    download: bool
    test: bool
    token: Optional[str] = Field(default=None)
    pytest_args: PytestArgs


class AgentRunnerClass(BaseModel):
    class_name: str
    class_path: str



import json
from difflib import SequenceMatcher
from typing import Any, Dict, List, Union

import pytest
from deepdiff import DeepDiff

AllowedJSON = Union[Dict[str, Any], List[str], List[Dict[str, Any]], str]

Result = Dict[str, Any]


def validate_field_match(expected: Result, actual: Result, field: str) -> None:
    expected_value = expected.get(field, None)
    actual_value = actual.get(field, None)

    # Treat empty string as none
    if actual_value == "":
        actual_value = None

    if not check_match(expected_value, actual_value):
        pytest.fail(f"{expected_value} != {actual_value}")


def validate_json_match(expected: AllowedJSON, actual: AllowedJSON) -> None:
    if isinstance(expected, Dict) and isinstance(actual, Dict):
        expected = format_new_lines(expected)
        actual = format_new_lines(actual)

        # TODO: Pass in schema in the backend and handle this OUTSIDE of tests
        # Adding missing keys in actual with None if they are expected to be None
        for key, value in expected.items():
            if value is None and key not in actual:
                actual[key] = None

    diff = DeepDiff(
        expected,
        actual,
        ignore_order=True,
        report_repetition=True,
    )
    if diff:
        # Pretty print both expected and actual results
        pretty_expected = json.dumps(expected, indent=4)
        pretty_actual = json.dumps(actual, indent=4)

        diff_msg = f"Actual: {pretty_actual}\nExpected: {pretty_expected}"
        pytest.fail(f"JSONEval mismatch!\n{diff_msg}")


def validate_end_url_match(expected: str, actual: str) -> None:
    if actual != expected:
        diff_msg = f"Actual URL:\t{actual}\nExpected URL:\t{expected}"
        pytest.fail(f"URLEval mismatch!\n{diff_msg}")


def format_new_lines(d: Result) -> Result:
    """Recursively replace newlines in strings with spaces."""
    new_dict: Result = {}
    for k, v in d.items():
        if isinstance(v, dict):
            new_dict[k] = format_new_lines(v)
        elif isinstance(v, str):
            new_dict[k] = v.replace("\n", " ")
        else:
            new_dict[k] = v
    return new_dict


def sanitize_string(input_str: str) -> str:
    return "".join(char for char in input_str if char.isalnum()).lower()


def is_string_similar(actual: str, expected: str, tolerance: int = 2) -> bool:
    if tolerance == 0:
        return actual == expected

    sanitized_actual = sanitize_string(actual)
    sanitized_expected = sanitize_string(expected)

    # Check if alphanumeric content matches
    if sanitized_actual != sanitized_expected:
        return False

    diff_count = native_count_differences(actual, expected)
    if diff_count <= tolerance:
        return True

    return SequenceMatcher(None, actual, expected).ratio() >= 0.8


def native_count_differences(actual: str, expected: str) -> int:
    non_alnum_actual = "".join(char for char in actual if not char.isalnum())
    non_alnum_expected = "".join(char for char in expected if not char.isalnum())
    # Compare the sequence of non-alphanumeric characters with a tolerance for
    # additional/missing characters
    diff_count = 0
    for char1, char2 in zip(non_alnum_actual, non_alnum_expected):
        if char1 != char2:
            diff_count += 1
    # Account for length difference if one sequence is longer than the other
    length_diff = abs(len(non_alnum_actual) - len(non_alnum_expected))
    diff_count += length_diff
    return diff_count


def check_match(expected_value: Any, actual_value: Any) -> bool:
    if isinstance(expected_value, str) and isinstance(actual_value, str):
        return is_string_similar(expected_value, actual_value)
    else:
        return expected_value == actual_value



from typing import Dict
from urllib.parse import urlparse

from bananalyzer import Example
from bananalyzer.data.schemas import Eval
from bananalyzer.runner.runner import BananalyzerTest


class PytestTestGenerator:
    def __init__(self) -> None:
        self._classnames: Dict[str, int] = {}

    def generate_test(self, example: Example) -> BananalyzerTest:
        return BananalyzerTest(
            code=f"""
@pytest.mark.asyncio
class {self._generate_class_name(example)}:

    @classmethod
    def setup_class(cls):
        cls.example = get_example_by_url("{example.url}")


    @pytest_asyncio.fixture(scope="class")
    async def result(self, page, agent):
        yield await agent.run(page, self.example)

    {"".join(self._generate_eval_test(eval_, i) for i, eval_ in enumerate(example.evals))}
""",
            example=example,
        )

    def _generate_eval_test(self, eval_: Eval, i: int) -> str:
        if eval_.type == "json_match" and isinstance(eval_.expected, dict):
            return f"""
    @pytest.mark.parametrize("key", {list(eval_.expected.keys())})
    async def test_match_field(self, key, result) -> None:
        self.example.evals[{i}].eval_results(None, result, field=key)

"""
        return f"""
    async def test_{eval_.type}(self, page, result) -> None:
        self.example.evals[{i}].eval_results(page, result)

"""

    def _generate_class_name(self, example: Example) -> str:
        domain = urlparse(example.url).netloc
        domain = domain.replace(".", "_")
        domain = domain.replace("-", "_")
        if domain.startswith("www_"):
            domain = domain[4:]

        domain = "".join([part.capitalize() for part in domain.split("_")])

        key = f"{example.type.capitalize()}{domain}"
        self._classnames[key] = self._classnames.get(key, -1) + 1
        suffix = "" if not self._classnames[key] else f"{self._classnames[key] + 1}"
        return f"Test{key}{suffix}"



from abc import ABC, abstractmethod
from pathlib import Path

from bananalyzer.data.examples import get_examples_path
from bananalyzer.data.schemas import Example


class WebsiteResponder(ABC):
    """
    Abstract class dealing with returning a website response
    """

    @abstractmethod
    def get_url(self, example: Example) -> str:
        """
        Get the website response
        """
        raise NotImplementedError("get_url not implemented")


class MHTMLWebsiteResponder(WebsiteResponder):
    """
    Return local MHTML files as the response. MHTML is a format that embeds all resources in a single file.
    Ideally this would be hosted by a webserver but browsers like chrome don't support MHTML
    """

    def __init__(self, data_path: Path):
        super().__init__()
        self.data_path = data_path

    def get_url(self, example: Example) -> str:
        mhtml_path = self.data_path / example.id / "index.mhtml"
        return f"file://{mhtml_path.as_posix()}"


class HostedWebsiteResponder(WebsiteResponder):
    """
    Some stable environments are already hosted on a stable server (WebArena) so we can just return the URL directly.
    """

    def get_url(self, example: Example) -> str:
        return example.url


def get_website_responder(example: Example) -> WebsiteResponder:
    if example.source == "mhtml":
        return MHTMLWebsiteResponder(get_examples_path())
    elif example.source == "hosted":
        return HostedWebsiteResponder()
    else:
        raise ValueError(f"Unknown example source {example.source}")



# Separate banana-lyzer args from pytest args
# Look for an instance of Banana-lyzer in the current directory
# If it doesn't exist, error
import argparse
import ast
import importlib.util
import os
import sys
from pathlib import Path
from typing import List
from urllib.parse import urlparse

import requests

from bananalyzer import AgentRunner
from bananalyzer.data.examples import (
    download_examples,
    get_test_examples,
    get_training_examples,
)
from bananalyzer.runner.generator import PytestTestGenerator
from bananalyzer.runner.runner import run_tests
from bananalyzer.schema import AgentRunnerClass, Args, PytestArgs


def print_intro() -> None:
    # https://www.asciiart.eu/food-and-drinks/bananas
    print(
        r"""
//\
V  \
 \  \_
  \,'.`-.
   |\ `. `.
   ( \  `. `-.                        _,.-:\
    \ \   `.  `-._             __..--' ,-';/
     \ `.   `-.   `-..___..---'   _.--' ,'/
      `. `.    `-._        __..--'    ,' /
        `. `-_     ``--..''       _.-' ,'
          `-_ `-.___        __,--'   ,'
             `-.__  `----'''    __.-'
                  `--..____..--'
"""
    )

    print("Bananalyzing... 🍌")


def parse_args() -> Args:
    file_name = "bananalyzer-agent.py"
    parser = argparse.ArgumentParser(
        description=f"Run the agent inside a bananalyzer agent definition file "
        f"against the benchmark",
    )
    parser.add_argument("path", type=str, help=f"Path to the {file_name} file")
    parser.add_argument(
        "--headless", action="store_true", help=f"Whether to run headless or not"
    )
    parser.add_argument(
        "-s",
        "--s",
        action="store_true",
        help="Shortcut for --capture=no in pytest. Will print stdout and stderr",
    )
    parser.add_argument(
        "-id",
        "--id",
        type=str,
        default=None,
        help="Filter tests by id",
    )
    parser.add_argument(
        "-d",
        "--domain",
        type=str,
        default=None,
        help="Filter tests by a particular URL domain",
    )
    parser.add_argument(
        "-i",
        "--intent",
        type=str,
        default=None,
        help="Filter tests by a particular intent",
    )
    parser.add_argument(
        "-c",
        "--category",
        type=str,
        default=None,
        help="Filter tests by a particular category",
    )
    parser.add_argument(
        "--subcategory",
        type=str,
        default=None,
        help="Filter tests by a particular category",
    )
    parser.add_argument(
        "-n",
        "--n",
        type=str,
        default=None,
        help="Number of test workers to use. The default is 1",
    )
    parser.add_argument(
        "-skip",
        "--skip",
        type=lambda s: s.split(","),
        default=[],
        help="A list of ids to skip tests on, separated by commas",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="Will decrease the verbosity of pytest. By default we run with the `--v` pytest param.",
    )
    parser.add_argument(
        "--single_browser_instance",
        action="store_true",
        help="Run tests in a single browser instance as opposed to creating a browser "
        "instance per test. This is faster but less reliable as test contexts can "
        "occasionally bleed into each other, causing tests to fail",
    )
    parser.add_argument(
        "--type",
        type=str,
        default=None,
        help="Filter tests by a particular type",
    )
    parser.add_argument(
        "--download",
        action="store_true",
        help="Will re-download training and test examples",
    )
    parser.add_argument(
        "--test",
        action="store_true",
        help="Use test set examples instead of training set examples",
    )
    parser.add_argument(
        "--token",
        type=str,
        default=None,
        help="The token to use when uploading results to the dashboard",
    )

    args = parser.parse_args()

    file_name = os.path.basename(args.path)
    if file_name != file_name:
        raise RuntimeError(f"The provided file name must be {file_name}")

    return Args(
        path=args.path,
        headless=args.headless,
        intent=args.intent,
        id=args.id,
        domain=args.domain,
        category=args.category,
        subcategory=args.subcategory,
        skip=args.skip,
        single_browser_instance=args.single_browser_instance,
        type=args.type,
        test=args.test,
        download=args.download,
        token=args.token,
        pytest_args=PytestArgs(
            s=args.s,
            n=args.n,
            q=args.quiet,
        ),
    )


def find_agents(file_path: Path) -> List[AgentRunnerClass]:
    with open(file_path, "r") as source:
        node = ast.parse(source.read())

    runners: List[AgentRunnerClass] = []
    for clazz in [n for n in node.body if isinstance(n, ast.ClassDef)]:
        if "AgentRunner" in [getattr(base, "id", "") for base in clazz.bases]:
            runners.append(
                AgentRunnerClass(
                    class_name=clazz.name,
                    class_path=str(file_path),
                )
            )

    return runners


def load_agent_from_path(path: Path) -> AgentRunnerClass:
    if path.is_dir():
        files = [p for p in path.glob("**/*.py") if "venv" not in p.parts]
    else:
        files = [path]

    runners: List[AgentRunnerClass] = []
    for file in files:
        runners.extend(find_agents(file))

    if len(runners) == 0:
        raise RuntimeError(f"Could not find any agent runners in {path}")

    if len(runners) > 1:
        raise RuntimeError(f"Found multiple agent runners in {path}")

    runner = runners[0]
    runner_file = Path(runner.class_path)
    module_name = path.stem

    spec = importlib.util.spec_from_file_location(module_name, runner_file)
    if spec is None or spec.loader is None:
        raise ImportError(f"Cannot load module from path {runner_file}")

    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)

    agent = getattr(module, runner.class_name)()
    if not isinstance(agent, AgentRunner):
        raise TypeError("User defined agent is is not an instance of AgentRunner")

    return runner


def main() -> int:
    """
    Load the agent from the provided path and run it against the benchmark

    Note that pytest creates a new global context when running tests.
    Because of this, we first load the agent and validate that it is of the correct type here.
    Then we pass the path to the agent runner and let it load it within the pytest context.
    Note your AgentRunner must be concurrency safe.
    """
    print_intro()

    # Load the agent
    args = parse_args()
    agent = load_agent_from_path(Path(args.path))

    print(f"Loaded agent {agent.class_name} from {agent.class_name}")

    if args.download:
        print("##################################################")
        print("# Downloading examples, this may take a while... #")
        print("##################################################")
        download_examples()

    # Filter examples based on args
    filtered_examples = get_test_examples() if args.test else get_training_examples()

    if args.id:
        filtered_examples = [
            example for example in filtered_examples if example.id == args.id
        ]

    if args.intent:
        filtered_examples = [
            example for example in filtered_examples if example.type == args.intent
        ]
    if args.domain:
        filtered_examples = [
            example
            for example in filtered_examples
            if ".".join(urlparse(example.url).netloc.split(".")[-2:]) == args.domain
        ]
    if args.category:
        filtered_examples = [
            example
            for example in filtered_examples
            if example.category == args.category
        ]
    if args.skip:
        filtered_examples = [
            example for example in filtered_examples if example.id not in args.skip
        ]
    if args.type:
        filtered_examples = [
            example for example in filtered_examples if example.type == args.type
        ]
    if args.subcategory:
        filtered_examples = [
            example
            for example in filtered_examples
            if example.subcategory in args.subcategory
        ]

    # Test we actually have tests to run
    if len(filtered_examples) == 0:
        print()
        print("=======================================================================")
        print("🍌 No tests to run. Please ensure your filter parameters are correct 🍌")
        print("=======================================================================")
        return 0

    # Load the desired tests
    generator = PytestTestGenerator()
    tests = [generator.generate_test(example) for example in filtered_examples]

    # Run the tests
    exit_code, report_path = run_tests(
        tests, agent, args.pytest_args, args.headless, args.single_browser_instance
    )

    if args.token:
        with open(report_path, "r") as fp:
            res = requests.post(
                "http://localhost:3000/api/rest/upload",
                headers={
                    "Authorization": f"Bearer {args.token}",
                    "Content-Type": "text/plain",
                },
                data=fp.read(),
            )
            res.raise_for_status()

    return exit_code


if __name__ == "__main__":
    main()



from typing import Dict, List, Optional, Type

from pydantic import BaseModel, Field, HttpUrl

from bananalyzer.data.schemas import FetchId

"""
This file contains mapping of fetch_id to fetch schema to avoid duplicate schemas in examples.json
"""


class ContactSchema(BaseModel):
    name: str = Field(
        description="name of the location (not the hospital system name) *precisely* as it is written on the page (do not edit it, add text, or combine names)"
    )
    address: str = Field(
        description="complete address of the location including street, city, state, and ZIP",
    )
    phone: str = Field(
        description="phone number of the location (only include the number but retain its formatting)",
    )
    fax: Optional[str] = Field(
        description="fax number of the location (only include the number but retain its formatting)",
        default=None,
    )
    type: Optional[str] = Field(
        description="the type of location: Neurosurgery, MRI Services, etc. (not all locations will have a type available on the page)",
        default=None,
    )


class JobPostingSchema(BaseModel):
    job_id: str
    job_title: str
    job_category: str
    date_posted: str
    location: str
    job_description: str
    roles_and_responsibilities: str
    qualifications: str
    preferred_qualifications: str
    benefits: str
    salary: str


class Specification(BaseModel):
    label: str
    value: str


class Document(BaseModel):
    url: str
    filename: str


class ManufacturingCommerceSchema(BaseModel):
    mpn: str
    alias_mpns: list[str] = Field(description="Other MPNs that this part is known by")
    manufacturer: str
    classifications: list[str]
    description: str
    hero_image: str
    series: str
    lifecycle_status: str
    country_of_origin: str
    aecq_status: str
    reach_status: str
    rohs_status: str
    export_control_class_number: str
    packaging: str
    power_rating: str
    voltage_rating: str
    mount_type: str
    moisture_sensitivity_level: str
    tolerance: str
    inductance: str
    capacitance: str
    resistance: str
    min_operating_temperature: str
    max_operating_temperature: str
    leadfree: str
    termination_type: str
    num_terminations: int
    specs: list[Specification]
    product_change_notification_documents: list[Document]
    reach_compliance_documents: list[Document]
    rohs_compliance_documents: list[Document]
    datasheets: list[Document]
    specsheets: list[Document]
    suggested_alternative_mpns: list[str]


class ForumSchema(BaseModel):
    author: str
    title: str
    post_date: str
    content: str
    up_votes: int
    down_votes: int
    views: int
    num_comments: int


class AttorneyExperience(BaseModel):
    title: str = Field(description="Title at the previous law firm")
    firm: str = Field(description="Name of the law firm")
    dates: str = Field(description="Duration of employment at the law firm")


class AttorneyBarAdmission(BaseModel):
    state: str = Field(description="State of bar admission")
    year: str = Field(description="Year of bar admission")
    country: Optional[str] = Field(
        description="Country of bar admission", default="USA"
    )


class AttorneyEducation(BaseModel):
    school: str = Field(description="Name of the educational institution")
    year: str = Field(description="Year of graduation")
    degree: Optional[str] = Field(description="Type of degree obtained", default=None)
    honors: Optional[str] = Field(
        description="Honors received during education", default=None
    )


class AttorneyAward(BaseModel):
    date: str = Field(description="Date or duration when the award was received")
    award: str = Field(description="Name of the award")
    url: Optional[HttpUrl] = Field(
        description="URL link to the award recognition", default=None
    )


class AttorneySchema(BaseModel):
    website: HttpUrl = Field(description="URL of the attorney's profile page")
    name: str = Field(description="Full name of the attorney")
    title: str = Field(
        description="Title of the attorney, such as Associate, Counsel, or Partner"
    )
    practice_areas_main: List[str] = Field(
        description="Primary practice areas of the attorney"
    )
    practice_areas_all: List[str] = Field(description="Complete list of practice areas")
    specialties: List[str] = Field(description="Specialized industry sectors")
    email: str = Field(description="Email address of the attorney")
    location: str = Field(description="Office location of the attorney")
    phone: Optional[str] = Field(
        description="Direct phone number of the attorney", default=None
    )
    bio: str = Field(description="Main bio description of the attorney")
    experience: List[AttorneyExperience] = Field(
        description="Past work history at other law firms"
    )
    matters: str = Field(description="List of past cases and work done by the attorney")
    bar_admissions: List[AttorneyBarAdmission] = Field(
        description="Bar admissions of the attorney"
    )
    law_school: List[AttorneyEducation] = Field(
        description="Law school information of the attorney"
    )
    other_schools: List[AttorneyEducation] = Field(
        description="Other education details of the attorney"
    )
    awards: List[AttorneyAward] = Field(
        description="Awards and recognitions received by the attorney"
    )
    pdf_url: Optional[HttpUrl] = Field(
        description="Link to a PDF bio of the attorney", default=None
    )
    photo_url: HttpUrl = Field(description="Link to the photo of the attorney")
    news: List[HttpUrl] = Field(
        description="Links to news articles involving the attorney", default=[]
    )


class AttorneyJobPostingSchema(BaseModel):
    website: HttpUrl = Field(
        description="URL of the job listing. May or may not be unique to this job."
    )
    tier: str = Field(
        description="Categorize into: Associate (general), Junior Associate, Mid-Level Associate, Senior Associate, Partner, Other (catch all if unsure)."
    )
    department: Optional[str] = Field(
        description="Department, only if explicitly stated.", default=None
    )
    title: str = Field(
        description="Job title. Remove location but keep everything else."
    )
    description: str = Field(description="Job description.")
    locations: List[str] = Field(
        description="Offices/cities/locations where this job is being offered."
    )
    salary_range: str = Field(
        description="Pay range specifically stated in the listing."
    )
    experience: str = Field(
        description="Required experience specifically stated in the listing."
    )
    full: str = Field(
        description="Complete information related to this job including HTML content for links."
    )


def get_fetch_schema(fetch_id: FetchId) -> Type[BaseModel]:
    fetch_schemas: Dict[str, Type[BaseModel]] = {
        "contact": ContactSchema,
        "job_posting": JobPostingSchema,
        "manufacturing_commerce": ManufacturingCommerceSchema,
        "forum": ForumSchema,
        "attorney": AttorneySchema,
        "attorney_job_listing": AttorneyJobPostingSchema,
    }

    if fetch_id not in fetch_schemas:
        raise ValueError(f"Invalid fetch_id: {fetch_id}")

    return fetch_schemas[fetch_id]



from bananalyzer.data.examples import get_all_examples
from bananalyzer.data.schemas import Example, GoalType
from bananalyzer.runner.agent_runner import AgentRunner
from bananalyzer.runner.evals import check_match

__all__ = ["AgentRunner", "GoalType", "Example", "get_all_examples", "check_match"]



import asyncio
from random import random
from typing import Any, cast

from playwright.async_api import Page

from bananalyzer.data.schemas import Example
from bananalyzer.runner.agent_runner import AgentResult, AgentRunner


class NullAgentRunner(AgentRunner):
    """
    A test agent class that just returns an empty string
    """

    RANDOM_FAILURE_RATE = 0  # You can change this to 0.5 to see random failures

    async def run(
        self,
        page: Page,
        example: Example,
    ) -> AgentResult:
        print(f"Testing {example.get_static_url()}")
        await page.goto(example.get_static_url())
        await asyncio.sleep(0.2)

        print(f"Done testing {example.get_static_url()}")

        # Ensure page is correct
        if example.evals[0].type == "end_url_match" and isinstance(
            example.evals[0].expected, str
        ):
            await page.goto(example.evals[0].expected)
            return example.evals[0].expected

        if example.type == "links" or example.type == "links_fetch":
            return example.evals[0].expected

        copy = cast(dict[str, Any], example.evals[0].expected).copy()
        for key, value in copy.items():
            if random() < self.RANDOM_FAILURE_RATE:
                copy[key] = "random"

        return copy



from typing import Any

import pytest
from _pytest.terminal import TerminalReporter
from tabulate import tabulate


def print_field_data(
    terminalreporter: TerminalReporter, results: dict[str, dict[bool, int]]
) -> None:
    headers = "Group", "Passed", "Failed", "Correct %", "Perfect"

    table_data = []
    total_passed = 0
    total_failed = 0
    perfect_count = 0

    for field, values in results.items():
        passed = values.get(True, 0)
        failed = values.get(False, 0)
        total_passed += passed
        total_failed += failed
        is_perfect = failed == 0
        perfect_count += int(is_perfect)
        percentage = (
            f"{passed / (passed + failed) * 100:.2f}%"
            if (passed + failed) > 0
            else "N/A"
        )
        perfect = is_perfect and "✅" or "❌"
        table_data.append([field, passed, failed, percentage, perfect])

    table_data.sort(key=lambda row: row[3], reverse=True)  # type: ignore

    # Calculate the percentage for the total row
    total_percentage = (
        f"{total_passed / (total_passed + total_failed) * 100:.2f}%"
        if (total_passed + total_failed) > 0
        else "N/A"
    )
    perfect_percentage = (
        f"{perfect_count / len(results) * 100:.2f}%" if results else "N/A"
    )

    # Add a totals row
    table_data.append(["-" * len(header) for header in headers])
    table_data.append(
        ["Total", total_passed, total_failed, total_percentage, perfect_percentage]
    )

    # Create a table using tabulate
    table = tabulate(table_data, headers=headers, tablefmt="psql")

    # Print the table
    terminalreporter.write_line(table)
    terminalreporter.write_line("\n")


class BananalyzerPytestPlugin:
    @pytest.hookimpl(trylast=True)
    def pytest_terminal_summary(
        self, terminalreporter: TerminalReporter, *args: Any, **kwargs: Any
    ) -> None:
        terminalreporter.section("Bananalyzer Results")

        results: dict[str, dict[str, dict[bool, int]]] = {}
        for test_result in (
            terminalreporter.stats.get("passed", [])
            + terminalreporter.stats.get("failed", [])
            + terminalreporter.stats.get("error", [])
        ):
            if not hasattr(test_result, "user_properties"):
                continue
            for key, value in test_result.user_properties:
                result_property = results.setdefault(key, {})
                result_property_value = result_property.setdefault(value, {})
                result_property_value[test_result.passed] = (
                    result_property_value.get(test_result.passed, 0) + 1
                )

        total_passed = len(terminalreporter.stats.get("passed", []))
        total_failed = len(terminalreporter.stats.get("failed", [])) + len(
            terminalreporter.stats.get("error", [])
        )
        total_tests = total_passed + total_failed

        if "field" in results:
            terminalreporter.write_line("Field Results:")
            print_field_data(terminalreporter, results["field"])

        if "class" in results:
            terminalreporter.write_line("Class Results:")
            print_field_data(terminalreporter, results["class"])

        table_data = {
            "Total Tests": total_tests,
            "Tests Passed": total_passed,
            "Tests Failed": total_failed,
            "Percent Passed": f"{total_passed / total_tests * 100:.2f}%",
        }

        terminalreporter.write_line("Summary:")
        terminalreporter.write_line(tabulate(table_data.items(), tablefmt="psql"))

    # noinspection PyBroadException
    @pytest.fixture(autouse=True)
    def add_user_properties(self, record_property, request) -> None:  # type: ignore
        for key, accessor in {
            "class": lambda *_: request.cls.__name__,
            "field": lambda *_: request.node.callspec.params.get("key"),
        }.items():
            try:
                record_property(key, accessor())
            except Exception:
                pass



from typing import Any, Dict, List, Literal, Optional, Union

import pydantic
from playwright.async_api import Page
from pydantic import BaseModel, Field, model_validator

from bananalyzer.runner.evals import (
    AllowedJSON,
    validate_end_url_match,
    validate_field_match,
    validate_json_match,
)

GoalType = Literal[
    "fetch",  # Scrape specific JSON information from a single page. Does not require navigation
    "links",  # Scrape all detail page links from a single listing page
    "links_fetch",  # Scrape all detail page links from a single listing page along with JSON information
    "click",  # Make a single click on a page
    "navigate",  # Travel to a new page
    "search",  # Search for the answer to a specific query
    "multiple",  # Perform multiple intents
]


class Eval(BaseModel):
    """
    Base class for all evals. Evals are used to determine if an action or result is correct
    """

    type: Literal["json_match", "end_url_match"] = "json_match"
    expected: AllowedJSON

    def eval_action(self, action: str) -> bool:
        """
        We don't care for action level evals at the moment
        """
        raise NotImplementedError("eval_action not implemented")

    def eval_results(
        self, page: Page, result: Dict[str, Any], field: Optional[str] = None
    ) -> None:
        if (
            self.type == "json_match"
            and field is not None
            and type(self.expected) is dict
        ):
            return validate_field_match(self.expected, result, field)

        if (
            self.type == "json_match"
            and type(self.expected) is dict
            or type(self.expected) is list
        ):
            return validate_json_match(self.expected, result)

        if self.type == "end_url_match" and type(self.expected) is str:
            return validate_end_url_match(self.expected, page.url)

        raise NotImplementedError("No evaluation type implemented")


FetchId = Literal[
    "job_posting",
    "manufacturing_commerce",
    "contact",
    "forum",
    "attorney",
    "attorney_job_listing",
]


class Example(BaseModel):
    id: str
    url: str
    source: Literal["mhtml", "hosted", "url"] = Field(
        description="Source of the website"
    )
    category: str = Field(description="Category of the website")
    subcategory: str = Field(description="Subcategory of the website")
    type: GoalType = Field(
        description="The high level goal intent the agent is aiming to do"
    )
    goal: Optional[Union[str, Dict[str, Any]]] = Field(
        description="The goal of the agent for this specific example"
    )
    fetch_id: Optional[FetchId] = Field(
        default=None,
        description="If it is a fetch type, we can infer the goal based on this id to avoid large schemas in json",
    )
    evals: List[Eval] = Field(
        description="Various evaluations to test for within the example"
    )

    def get_static_url(self) -> str:
        from bananalyzer.runner.website_responder import get_website_responder

        return get_website_responder(self).get_url(self)

    @model_validator(mode="before")
    def set_goal_if_fetch_id_provided(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        from bananalyzer.data.fetch_schemas import get_fetch_schema

        goal_type = values.get("type")
        if goal_type != "fetch":
            return values

        fetch_id: Optional[FetchId] = values.get("fetch_id")
        goal = values.get("goal")

        if fetch_id is not None and goal is not None:
            raise ValueError("fetch_id and goal cannot both be provided")

        if fetch_id is None and goal is not None:
            return values

        if fetch_id is None:
            raise ValueError("fetch_id must be provided if goal is not provided")

        values["goal"] = get_fetch_schema(fetch_id).model_fields
        return values



import json
import os
import shutil
import subprocess
from pathlib import Path
from typing import List

from bananalyzer.data.schemas import Example

local_examples_path: Path = Path(__file__).resolve().parent.parent.parent / "static"
downloaded_examples_path = Path.home() / ".bananalyzer_data"
train_examples_name = "examples.json"
test_examples_name = "test_examples.json"


def are_examples_available(path: Path) -> bool:
    if not path.exists():
        return False

    return (path / train_examples_name).exists() and (
        path / test_examples_name
    ).exists()


def get_examples_path() -> Path:
    if are_examples_available(local_examples_path):
        print("### Using local examples! ###")
        return local_examples_path
    elif are_examples_available(downloaded_examples_path):
        return downloaded_examples_path
    else:
        raise FileNotFoundError(
            "No examples download. Re-run with `--download` to download example data via git."
        )


def convert_to_crlf(file_path: Path) -> None:
    """
    Git automatically replaces CRLF endings with LF either on push or pull
    This causes the MHTML to become invalid and result in a blank page
    As a result, we must manually replace LF with CRLF
    """
    with open(file_path, "rb") as file:
        lines = file.readlines()

    with open(file_path, "wb") as file:
        for line in lines:
            if line.endswith(b"\n") and not line.endswith(b"\r\n"):
                line = line.replace(b"\n", b"\r\n")
            file.write(line)


def download_examples() -> None:
    """
    Downloads the repo via git and places contents of the `/static` data directory in ~/.bananalyzer_data
    """
    repo_url = "https://github.com/reworkd/bananalyzer.git"
    branch = "main"
    data_folder_name = "static/"

    try:
        subprocess.run(
            ["git", "clone", "-b", branch, repo_url, "repo_temp"], check=True
        )

        data_folder_path = Path("repo_temp") / data_folder_name
        if not data_folder_path.exists():
            raise FileNotFoundError(
                f"The folder '{data_folder_name}' does not exist in the repository."
            )

        downloaded_examples_path.mkdir(parents=True, exist_ok=True)
        for item in downloaded_examples_path.iterdir():
            if item.is_dir():
                shutil.rmtree(item)
            else:
                item.unlink()

        for item in data_folder_path.iterdir():
            target_path = shutil.move(str(item), downloaded_examples_path)
            for root, dirs, files in os.walk(target_path):
                for file in files:
                    convert_to_crlf(Path(root) / file)

    finally:
        print("Cleaning up repo...")
        shutil.rmtree("repo_temp", ignore_errors=True)


def load_examples_at_path(path: Path, examples_json_file_name: str) -> List[Example]:
    examples_json_path = path / examples_json_file_name
    try:
        with open(examples_json_path, "r") as file:
            example_jsons = json.load(file)
    except json.JSONDecodeError as e:
        print(f"Error decoding JSON from {examples_json_path}: {e}")
        raise
    except FileNotFoundError as e:
        print(f"Example file not found: {e}")
        raise

    return [Example(**example) for example in example_jsons]


def get_training_examples() -> List[Example]:
    return load_examples_at_path(get_examples_path(), train_examples_name)


def get_test_examples() -> List[Example]:
    return load_examples_at_path(get_examples_path(), test_examples_name)


def get_all_examples() -> List[Example]:
    return get_training_examples() + get_test_examples()


def get_all_example_urls() -> List[str]:
    return [example.url for example in get_all_examples()]


def get_example_by_url(url: str) -> Example:
    for example in get_all_examples():
        if example.url == url:
            return example
    raise ValueError(f"No example found with URL: {url}")
